<?php
/**
 * montonio/module-hire-purchase
 *
 * @category   montonio
 * @package    module-montonio-m1
 * @copyright  Copyright (c) 2021 Montonio Finance OÜ  (https://montonio.com/)
 * @license    https://www.gnu.org/licenses/gpl-3.0.txt
 */

//require JWT lib
require_once Mage::getBaseDir('lib') . DS . 'Montonio' . DS . 'JWT.php';

class Montonio_Gateway_Model_Method_Payments extends Mage_Payment_Model_Method_Abstract
{

    const METHOD_CODE = 'montonio_payments';
    const METHOD_ALL_COUNTRIES = 'ALL';

    protected $_formBlockType               = 'montonio/payments_form';
    protected $_infoBlockType               = 'montonio/payments_info';
    protected $_code                        = self::METHOD_CODE;
    protected $_allowCurrencyCode           = ['EUR'];
    protected $_canManageRecurringProfiles  = false;
    protected $_canUseInternal              = false;

    /**
     * @inheritdoc
     */
    public function getCode()
    {
        $virtualCode = $this->getData('code');
        if ($virtualCode && strstr($virtualCode, self::METHOD_CODE.'_')) {
            return $virtualCode;
        } else {
            return parent::getCode();
        }
    }

    /**
     * @inheritdoc
     */
    public function getConfigData($field, $storeId = null)
    {
        if (null === $storeId) {
            $storeId = $this->getStore();
        }

        //virtual code is sometimes needed but not on config calls
        $code = $this->getCode();

        if (strstr($code, 'montonio_payments_')) {
            $subCode     = str_replace('montonio_payments_', '', $code);
            if ($field == 'country') {
                $methods = $this->getAvailableMethods();
                if (isset($methods[$subCode])) {
                    return $methods[$subCode]['method_country'];
                }
            }
            if ($field == 'title') {
                $methods = $this->getAvailableMethods();
                if (isset($methods[$subCode])) {
                    return $methods[$subCode]['method_title'];
                }
            }
            if ($field == 'description') {
                $methods = $this->getAvailableMethods();
                if (isset($methods[$subCode])) {
                    return $methods[$subCode]['method_desc'];
                }
            }
            if ($field == 'logo') {
                $methods = $this->getAvailableMethods();
                if (isset($methods[$subCode])) {
                    return $methods[$subCode]['logo_url'];
                }
            }
            if ($field == 'sort_order') {

                $methods = $this->getAvailableMethods();
                if (isset($methods[$subCode])) {
                    return $methods[$subCode]['method_sort'];
                }
            }
            $code = 'montonio_payments';
        }

        $path = 'payment/'. $code.'/'.$field;
        return Mage::getStoreConfig($path, $storeId);
    }

    /**
     * @inheritdoc
     */
    public function getOrderPlaceRedirectUrl()
    {
        $url            = Mage::getUrl('checkout/cart');
        $environment    = $this->getConfigData('environment');
        try {
            $info   = $this->getInfoInstance();
            $data   = $info->getAdditionalInformation();
            $gwUrl  = $this->getConfigData($environment.'_application_url') . '/?payment_token=';

            if (!empty($data[self::METHOD_CODE.'_token'])) {
                $url = $gwUrl.$data[self::METHOD_CODE.'_token'];
            } else {
                throw new Mage_Core_Exception('no token for building gateway redirect url');
            }
        } catch (Exception $e) {
            Mage::logException($e);
        }
        return $url;
    }

    /**
     *
     * @param  $order
     * @param  $payment
     * @return array
     */
    public function prepPayload($order)
    {

        $data           = [];
        $environment    = $this->getConfigData('environment');
        $time           = Mage::getModel('core/date')->timestamp(time());
        $pubKey         = $this->getConfigData($environment.'_public_key');
        $privKey        = $this->getConfigData($environment.'_private_key');
        $apiUrl         = $this->getConfigData($environment.'_api_url');
        $storeId        = $order->getStoreId();

        /* system data */
        $data['exp']        = $time + (60 * 10);
        $data['access_key'] = $this->getConfigData($environment.'_public_key', $storeId);

        $returnUrl  = $this->getConfigData('return_url', $storeId);
        $hookUrl    = $this->getConfigData('return_hook_url', $storeId);

        $data['merchant_return_url'] = Mage::getUrl($returnUrl);
        $data['merchant_notification_url'] = Mage::getUrl($hookUrl);

        //if grouped we don't send this so we get payment choices on gateway
        $isGrouped = $this->getConfigData('grouped');
        if (!$isGrouped) {
            $code = $order->getPayment()->getMethod();
            if (strstr($code, 'montonio_payments_')) {
                $subCode = str_replace('montonio_payments_', '', $code);
                $data['preselected_aspsp'] = $subCode;
            }
        }

        $storeLocale = Mage::app()->getLocale()->getLocaleCode();

        //montonio only supports following locales
        $locales = [
            'en_US' => 'en_US',
            'et_EE' => 'et',
            'lt_LT' => 'lt',
            'ru_RU' => 'ru'
        ];

        $localeCode = (isset($locales[$storeLocale])) ? $locales[$storeLocale] : $storeLocale;

        $data['preselected_locale'] = $localeCode;

        /*order data*/
        $data['amount'] = (float)sprintf("%.2f", $order->getGrandTotal());
        $data['currency'] = $order->getOrderCurrency()->toString();
        $data['merchant_reference'] = $order->getIncrementId();
        $data['payment_information_unstructured'] = Mage::helper('montonio')
                            ->__('Payment for order "%s"', $order->getIncrementId());

        if ($order->getCustomerEmail()) {
            $data['checkout_email'] = $order->getCustomerEmail();
        }

        if ($order->getBillingAddress()->getFirstname()) {
            $data['checkout_first_name'] = $order->getBillingAddress()->getFirstname();
        }

        if ($order->getBillingAddress()->getLastname()) {
            $data['checkout_last_name'] = $order->getBillingAddress()->getLastname();
        }

        if ($order->getBillingAddress()->getTelephone()) {
            $data['checkout_phone_number'] =  $order->getBillingAddress()->getTelephone();
        }

        $privateKey = $this->getConfigData($environment.'_private_key', $storeId);
        $applicationUrl= $this->getConfigData($environment.'_application_url', $storeId) . '/?payment_token=';
        $jwtToken = $this->generateSignature($data, $privateKey);

        $data['token'] = $jwtToken;

        return $data;
    }

    /**
     *
     * @param mixed $data
     * @param string $key
     * @return string
     */
    public function generateSignature($data, $key, $jsonEncode = false)
    {
        if ($jsonEncode) {
            $data = json_encode($data);
        }
        return JWT::encode($data, $key);
    }

    /**
     *
     * @param string $data
     * @param string $key
     * @return array|string
     */
    public function decodeSignature($data, $key)
    {
        return JWT::decode($data, $key);
    }

    /**
     *
     * @param array $data
     * @param string $accesKey
     * @return boolean
     */
    public function validateSignedData($data, $accesKey)
    {

        $requiredKeys = explode(',',$this->getConfigData('validate_fields'));
        $time   = Mage::getModel('core/date')->timestamp(time());
        $data = (array)$data;

        foreach($requiredKeys as $k => $v) {
            if(!isset($data[$v])) {
                Mage::throwException(Mage::helper('montonio')->__('Signature validation failed on missing field in message: "%s"', $v));
            }
        }

        if ($data['access_key'] !== $accesKey) {
            Mage::throwException(Mage::helper('montonio')->__('Signature validation failed on "%s" validation', 'access_key'));

        }
        if ($data['exp'] < $time) {
            Mage::throwException(Mage::helper('montonio')->__('Signature validation failed on "%s" validation', 'exp'));
        }

        return true;
    }

    /**
     *
     * @param string $token
     * @param string $storeId
     * @return array
     */
    public function validateIncomingData($token, $storeId = null)
    {
        $data           = [];
        $environment    = $this->getConfigData('environment');
        $privateKey     = $this->getConfigData($environment.'_private_key', $storeId);
        $accessKey      = $this->getConfigData($environment.'_public_key', $storeId);
        $data           = (Array)$this->decodeSignature($token, $privateKey);
        try {
            $this->validateSignedData($data, $accessKey);
        } catch (\Exception $e) {
            Mage::logException($e);
            $data = [];
        }

        return $data;
    }

    public function handleOrderStatus($data, $response)
    {

        $order  = Mage::getModel('sales/order')
                    ->loadByIncrementId($data['merchant_reference']);
        $helper = Mage::helper('montonio');


        if (!isset($order['increment_id']) ||
            $order['increment_id'] !== $data['merchant_reference'] ||
            $order['status'] !== 'pending'
        ) {
            $response['status'] = $order['status'];
            return $response;
        }

        $quote = Mage::getModel('sales/quote')
                         ->load($order['quote_id']);

        if ($data['status'] === 'finalized') {
            $newStatus = Mage_Sales_Model_Order::STATE_PROCESSING;
            $message = $helper->__('Payment accepted in payment gateway for "%s"', $data['amount']);
            $quote->setIsActive(0);

        } else {
            $newStatus = Mage_Sales_Model_Order::STATE_CANCELED;
            $message = $helper->__('Payment cancelled in payment gateway');
            $quote->setIsActive(1);
        }

        $transaction = Mage::getModel('core/resource_transaction');
        $transaction->addObject($quote);

        $order->setState($newStatus, true, $message, false);
        $order->setCanSendNewEmailFlag(true);

        $canCreateInvoice = $this->getConfigData('create_invoice', $order->getStoreId());

        //add invoice to order
        if (
            $order->getState() === Mage_Sales_Model_Order::STATE_PROCESSING &&
            $order->canInvoice() &&
            $canCreateInvoice
        ) {

            $invoice = Mage::getModel('sales/service_order', $order)->prepareInvoice();
            $invoice->setRequestedCaptureCase(Mage_Sales_Model_Order_Invoice::CAPTURE_OFFLINE);
            $invoice->register();
            $invoice->addComment($message,false,false);
            $invoice->getOrder()->setIsInProcess(true);

            $transaction->addObject($invoice);

            $order = $invoice->getOrder();

        }


        $transaction->addObject($order);
        $transaction->save();

        try {

            $canOrderEmail = $this->getConfigData('order_email_after', $order->getStoreId());

            if($order->getCanSendNewEmailFlag() && $canOrderEmail) {
                //for 1.9.* it turned to queue based method
                if(method_exists ($order, 'queueNewOrderEmail')) {
                    $order->queueNewOrderEmail();
                } else {
                    //for anything lower than 1.9 we need to use the old method
                    if(method_exists ($order, 'sendNewOrderEmail')) {
                        $order->sendNewOrderEmail();
                    }
                }
            }

        } catch (Exception $e) {
            Mage::logException($e);
        }

        try {

            $canInvoiceEmail = $this->getConfigData('invoice_email_after', $order->getStoreId());

            if($canCreateInvoice && is_object($invoice) && $canInvoiceEmail) {
                if(method_exists ($invoice, 'sendEmail')) {
                    $invoice->sendEmail();
                }
            }

        } catch (Exception $e) {
            Mage::logException($e);
        }

        $response = [
            'success' => true,
            'error' => false,
            'status' => $newStatus,
            'message' => $message
        ];

        if ($newStatus === Mage_Sales_Model_Order::STATE_PROCESSING) {
            $response['data'] = [
                'order_increment_id' => $order['increment_id'],
                'order_id' => $order['entity_id'],
                'quote_id' => $order['quote_id'],
            ];
        }

        return $response;
    }

    /**
     * get list of available Montonio methods from config
     *
     * @return array
     */
    public function getAvailableMethods()
    {
        $methods = [];

        try {
            $values     = unserialize($this->getConfigData('available_methods'));
            $methods    = array_flip(array_keys($values));

            foreach ($methods as $key => $value) {
                if (is_array($values[$key])) {
                    $methods[$key] = $values[$key];
                }

                if (is_array($values[$key]) && !$values[$key]['method_enabled']) {
                    unset($methods[$key]);
                }
            }
        } catch (\Exception $e) {
            Mage::logException($e);
        }
        //if grouped we add this to the mix so its counted as available
        if ($this->getValue('grouped')) {
            //grouped is a imaginary code for grouped method
            $methods['grouped'] = (!empty($methods)) ? (int)count($methods) + 1 : 0;
        }

        return $methods;
    }

    /**
     * Executes curl to get list of banks from gateway
     *
     * @throws \Exception
     * @return array
     */
    public function aspspsCommand()
    {

        $response = [];
        $environment    = $this->getConfigData('environment');

        $time           = Mage::getModel('core/date')->timestamp(time());
        $pubKey         = $this->getConfigData($environment.'_public_key');
        $privKey        = $this->getConfigData($environment.'_private_key');
        $apiUrl         = $this->getConfigData($environment.'_api_url');

        if (empty($privKey)) {
            return $response;
        }

        $apiEndpoint    = $this->getConfigData('listofbanks_api_endpoint');
        $transferObject = new Varien_Object();
        $payload = [
            'access_key' => $pubKey,
            'iat'        => $time,
            'exp'        => $time + (60 * 60)
        ];

        try {
            $bearerJwtToken = JWT::encode($payload, $privKey);
        } catch (Exception $e) {
            Mage::logException($e);
            throw new \Exception($e->getMessage());
        }

        $transferObject
            ->setUri($apiUrl.$apiEndpoint)
            ->setHeaders(
                [
                    'Authorization: Bearer ' . $bearerJwtToken,
                    'Content-Type: application/json'
                ]
            )
            ->setMethod('GET');

        try {
            $http = new Varien_Http_Adapter_Curl();
            $config = [
                'timeout'       => 30,
                'verifypeer'    => 0,
                'verifyhost'    => 0,
                'header'        => true,
            ];
            $http->setConfig($config);
            $http->write(
                $transferObject->getMethod(),
                $transferObject->getUri(),
                '1.1',
                $transferObject->getHeaders()
            );
            $response = $http->read();
            $response = json_decode(
                Zend_Http_Response::extractBody($response),
                true
            );
            $http->close();

            if (isset($response['statusCode']) && $response['statusCode'] !== '200') {
                $message = Mage::helper('montonio')->
                    __('aspsps request status "%s".', $response['statusCode']);
                throw new \Exception($message);
            }

        } catch (Exception $e) {
            Mage::logException($e);
            throw new \Exception($e->getMessage());
        }

        return $response;
    }

    /**
     *
     * @inheritdoc
     */
    public function canUseForCountry($countryCode) {

        $result = false;

        if(
            $this->getConfigData('country') !== '' &&
            (
                $countryCode == $this->getConfigData('country') ||
                $this->getConfigData('country') === self::METHOD_ALL_COUNTRIES ||
                $this->getConfigData('allowall') == 1
            )
        ) {
            $result =  true;
        } else if ($this->getConfigData('country') == '') {
            $result =  parent::canUseForCountry($countryCode);
        }

        return $result;
    }
}
